/******************************************************************************
 falf.pl version 1.0 3/6/2005 (Pedro Cabalar)
******************************************************************************/ 

%----- operators -------------------------------------------------------------

% arithmetic operators have priority 400 (*,/,mod) or 500 (+,-)
% relational operators have priority 700
:-op(300,fy,'prev.').
:-op(300,fy,'goal.').
:-op(700,xfx,:=).
:-op(700,yfx,<=).
:-op(700,yfx,'!=').
:-op(710,fy,not).
:-op(710,fy,'not -').
:-op(720,yfx,and).
:-op(730,yfx,or).
:-op(1200,yfx,if).

%----- general dynamic predicates ----------------------------------------------

:- dynamic decl/3.
   % decl(Type,Name,Data)
   % Type=sort, Name, Data=set expression
   % Type=var, Name, Data=set expression
   % Type=attr, Name/Arity, Data=[AttrType,Domain,Range,DefValue]
   %    AttrType belongs to {action, static, event, fluent}
   %	Domain is a list of set expressions
   %	Range is a set expression
   %	DefValue can be a singleton value [V] or no value []
   % Type=rule, Number/Line, Data=[RuleType,HeadExpr,BodyExpr]
   %    RuleType belongs to {initial,static,dynamic,goal}
   %    
   
:- dynamic
   sortnames/1,	  	% list of all sort names
   staticnames/1.  	% list of all static fluents

decl(sort,boolean,[false,true]).

%----- other "modules" --------------------------------------------------------
:- consult('util.pl').
:- consult('display.pl').
:- consult('loadfile.pl').
:- consult('translate.pl').
:- consult('solutions.pl').

:- dynamic opt/1.	% Program options like, for instance, opt(debug)

%----- main ------------------------------------------------------------------

% retrieves command line arguments Args and calls main(Args)

% For script version
main_s :- unix(argv(Argv)),
        ( append(_PrologArgs, [--|AppArgs], Argv), !,
          main(AppArgs)
        ; main(Argv)).

% For compiled version
main_c :- unix(argv(Argv)),
	Argv=[_ExecutableName|Args],
        main(Args).

% if no arguments, print help
main([]):- !,
  header,nl,
  write('falf [options] filenames'),nl,
  write('  -a            display only action values (unless -v enabled). '),nl,
  write('  -b <backend>  chooses <backend> syntax: lparse (the default) or dlv.'),nl,
  write('  -c            enable concurrent actions.'),nl,
  write('  -l <number>   path length is iterated until finding some length'),nl,
  write('                with solution(s) or reaching <number> transitions.'),nl,
  write('  -l=<number>   path length is exactly <number> transitions.'),nl,
  write('  -n <number>   computes at most <number> stable models. Default <number> is 1.'),nl,
  write('                When <number> is 0, computes all stable models.'),nl,
  write('  -o <filename> output file.'),nl,
  write('  -q            supresses information header and messages.'),nl,
  write('  -t <filename> store result of translation in a given file.'),nl,
  write('  -v            verbose mode: shows sort extents and default facts.'),nl,
  write('                When using -t, includes comments in the translation.'),nl,
  (opt(nohalt),!; halt).



:- dynamic plength/1.   % current path length  
% otherwise, set options depending on arguments
main(Args):-
	checkoptions(Args,Args2),
	(opt(backend(B)),!; B=lparse,asserta(opt(backend(B))) ),
	(opt(nummodels(N)),!; N=1,asserta(opt(nummodels(N))) ),
	(opt(pathlength(PL)),!; PL=0,asserta(opt(pathlength(PL))) ),
	loadfiles(Args2),
	
        (opt(translation(File)),!; File='falf.tmp'),
	translate(File),
	( opt(iterate),!,
	  set_count(plength,0),
	  repeat, iterated_write_solutions(B,File,N,PL),!
	; (write_solutions(B,File,N,PL),!;true)
	),
	(opt(nohalt),!; halt).

iterated_write_solutions(B,File,N,PL):-
	plength(L),
	( L > PL, !
	; incr(plength,1),
	  writelist(['\nTrying path length ',L,'...']),nl,
	  write_solutions(B,File,N,L),!
	).

write_solutions(B,File,N,PL):-
	command_line(B,File,N,PL,Command),
	shell(Command,_),
	see('falf.2.tmp'),
	(opt(outfile(F)),!,tell(F);true),
	display_solutions(B,NSol),
	seen,
	(opt(outfile(F)),!,told;true),
	delete_file('falf.2.tmp'),
	writelist([NSol,' solutions']),nl,NSol>0,!.

command_line(lparse,File,N,PL,Command):-
	!,concat_atom(['lparse --true-negation -c pathlength=',PL,' ',
	                File,' | smodels ',
	    N,' > falf.2.tmp'],Command).
command_line(dlv,File,N,PL,Command):-
	!,(PL>1,!,concat_atom([' -N=',PL,' '],TXT); TXT=' '),
	concat_atom(['dlv -silent -n=',N,TXT,File,
	    ' > falf.2.tmp '],Command).
	
checkoptions([],[]):-!.
checkoptions([X|Xs],Ys):- 
	    option(X,Opt),!,asserta(opt(Opt)),checkoptions(Xs,Ys)
	  ; X='-o',!, Xs=[File|Zs],
	    asserta(opt(outfile(File))),checkoptions(Zs,Ys)
	  ; X='-b',!,Xs=[Backend|Zs],
	    asserta(opt(backend(Backend))),checkoptions(Zs,Ys)
	  ; X='-n',!,Xs=[N|Zs],
	    asserta(opt(nummodels(N))),checkoptions(Zs,Ys)
	  ; X='-t',!,Xs=[File|Zs],
	    asserta(opt(translation(File))),checkoptions(Zs,Ys)
	  ; X='-l',!,Xs=[N|Zs],asserta(opt(iterate)),
	    asserta(opt(pathlength(N))),checkoptions(Zs,Ys)
	  ; name(X,L),L=[0'-,0'l,0'=|LL],!,
	    name(N,LL),
	    asserta(opt(pathlength(N))),checkoptions(Xs,Ys).
	    
checkoptions([X|Xs],[X|Ys]):- checkoptions(Xs,Ys).

option('-a',only_actions).
option('-c',conc).
option('-q',quiet).
option('-v',debug).
option('--nohalt',nohalt).

header :-
  writelogo(1),nl,
  writelogo(2),
  write(' Functional Action Language Frontend 1.0'),nl,
  writelogo(3),
  write(' (c) 2005- Pedro Cabalar, 4/6/2005, Univ. of Corunna'),nl,
  writelogo(4),
  write(' http://www.dc.fi.udc.es/ai/~cabalar/fal/'),nl,
  writelogo(5),nl.

writelogo(N):-logo(X),nth1(N,X,L),write(L).

logo([
'============ ====================================================',
',-  _  | ,- ',
':- (_\\ | :- ' ,
'|        |  ',
'============ ===================================================='
]).

